/*
*  This program is free software: you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation, either version 3 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
*  Release date: 20/05/2015
*  Authors: Norbert Bus & Nabil Mustafa
*  
*  If you find this code useful, please cite it by referencing the following paper:
*  @article{BMB15, title={Illumination{Cut}}, author={Bus, Norbert and Mustafa, Nabil H. and Biri, Venceslas},
*           journal = {Computer Graphics Forum (Proceedings of Eurographics 2015)},year={2015}}
*/

#include "devices/device_singleray/clustering/icStatistics.h"

namespace embree
{

  Stats stats;

  void ColormapJet::mapToJet( float value, float& r, float& g, float& b )
  {
    // handling special case (hitting the environment map)
    if (value < 0.0001) 
    {
      r = g = b = 0.0;
      return;
    }
    // value is some number from 0 to sometimes even 10K
    // how to map it to the range [0,256] ?
    int index = log(value+1) * 25;
    index = index > 255 ? 255 : index;
    
    // lazy to code...
    float jet[256][3] = {
      {     0,         0,    0.5156},
      {     0,         0,    0.5313},
      {     0,         0,    0.5469},
      {     0,         0,    0.5625},
      {     0,         0,    0.5781},
      {     0,         0,    0.5938},
      {     0,         0,    0.6094},
      {     0,         0,    0.6250},
      {     0,         0,    0.6406},
      {     0,         0,    0.6563},
      {     0,         0,    0.6719},
      {     0,         0,    0.6875},
      {     0,         0,    0.7031},
      {     0,         0,    0.7188},
      {     0,         0,    0.7344},
      {     0,         0,    0.7500},
      {     0,         0,    0.7656},
      {     0,         0,    0.7813},
      {     0,         0,    0.7969},
      {     0,         0,    0.8125},
      {     0,         0,    0.8281},
      {     0,         0,    0.8438},
      {     0,         0,    0.8594},
      {     0,         0,    0.8750},
      {     0,         0,    0.8906},
      {     0,         0,    0.9063},
      {     0,         0,    0.9219},
      {     0,         0,    0.9375},
      {     0,         0,    0.9531},
      {     0,         0,    0.9688},
      {     0,         0,    0.9844},
      {     0,         0,    1.0000},
      {     0,    0.0156,    1.0000},
      {     0,    0.0313,    1.0000},
      {     0,    0.0469,    1.0000},
      {     0,    0.0625,    1.0000},
      {     0,    0.0781,    1.0000},
      {     0,    0.0938,    1.0000},
      {     0,    0.1094,    1.0000},
      {     0,    0.1250,    1.0000},
      {     0,    0.1406,    1.0000},
      {     0,    0.1563,    1.0000},
      {     0,    0.1719,    1.0000},
      {     0,    0.1875,    1.0000},
      {     0,    0.2031,    1.0000},
      {     0,    0.2188,    1.0000},
      {     0,    0.2344,    1.0000},
      {     0,    0.2500,    1.0000},
      {     0,    0.2656,    1.0000},
      {     0,    0.2813,    1.0000},
      {     0,    0.2969,    1.0000},
      {     0,    0.3125,    1.0000},
      {     0,    0.3281,    1.0000},
      {     0,    0.3438,    1.0000},
      {     0,    0.3594,    1.0000},
      {     0,    0.3750,    1.0000},
      {     0,    0.3906,    1.0000},
      {     0,    0.4063,    1.0000},
      {     0,    0.4219,    1.0000},
      {     0,    0.4375,    1.0000},
      {     0,    0.4531,    1.0000},
      {     0,    0.4688,    1.0000},
      {     0,    0.4844,    1.0000},
      {     0,    0.5000,    1.0000},
      {     0,    0.5156,    1.0000},
      {     0,    0.5313,    1.0000},
      {     0,    0.5469,    1.0000},
      {     0,    0.5625,    1.0000},
      {     0,    0.5781,    1.0000},
      {     0,    0.5938,    1.0000},
      {     0,    0.6094,    1.0000},
      {     0,    0.6250,    1.0000},
      {     0,    0.6406,    1.0000},
      {     0,    0.6563,    1.0000},
      {     0,    0.6719,    1.0000},
      {     0,    0.6875,    1.0000},
      {     0,    0.7031,    1.0000},
      {     0,    0.7188,    1.0000},
      {     0,    0.7344,    1.0000},
      {     0,    0.7500,    1.0000},
      {     0,    0.7656,    1.0000},
      {     0,    0.7813,    1.0000},
      {     0,    0.7969,    1.0000},
      {     0,    0.8125,    1.0000},
      {     0,    0.8281,    1.0000},
      {     0,    0.8438,    1.0000},
      {     0,    0.8594,    1.0000},
      {     0,    0.8750,    1.0000},
      {     0,    0.8906,    1.0000},
      {     0,    0.9063,    1.0000},
      {     0,    0.9219,    1.0000},
      {     0,    0.9375,    1.0000},
      {     0,    0.9531,    1.0000},
      {     0,    0.9688,    1.0000},
      {     0,    0.9844,    1.0000},
      {     0,    1.0000,    1.0000},
      {0.0156,    1.0000,    0.9844},
      {0.0313,    1.0000,    0.9688},
      {0.0469,    1.0000,    0.9531},
      {0.0625,    1.0000,    0.9375},
      {0.0781,    1.0000,    0.9219},
      {0.0938,    1.0000,    0.9063},
      {0.1094,    1.0000,    0.8906},
      {0.1250,    1.0000,    0.8750},
      {0.1406,    1.0000,    0.8594},
      {0.1563,    1.0000,    0.8438},
      {0.1719,    1.0000,    0.8281},
      {0.1875,    1.0000,    0.8125},
      {0.2031,    1.0000,    0.7969},
      {0.2188,    1.0000,    0.7813},
      {0.2344,    1.0000,    0.7656},
      {0.2500,    1.0000,    0.7500},
      {0.2656,    1.0000,    0.7344},
      {0.2813,    1.0000,    0.7188},
      {0.2969,    1.0000,    0.7031},
      {0.3125,    1.0000,    0.6875},
      {0.3281,    1.0000,    0.6719},
      {0.3438,    1.0000,    0.6563},
      {0.3594,    1.0000,    0.6406},
      {0.3750,    1.0000,    0.6250},
      {0.3906,    1.0000,    0.6094},
      {0.4063,    1.0000,    0.5938},
      {0.4219,    1.0000,    0.5781},
      {0.4375,    1.0000,    0.5625},
      {0.4531,    1.0000,    0.5469},
      {0.4688,    1.0000,    0.5313},
      {0.4844,    1.0000,    0.5156},
      {0.5000,    1.0000,    0.5000},
      {0.5156,    1.0000,    0.4844},
      {0.5313,    1.0000,    0.4688},
      {0.5469,    1.0000,    0.4531},
      {0.5625,    1.0000,    0.4375},
      {0.5781,    1.0000,    0.4219},
      {0.5938,    1.0000,    0.4063},
      {0.6094,    1.0000,    0.3906},
      {0.6250,    1.0000,    0.3750},
      {0.6406,    1.0000,    0.3594},
      {0.6563,    1.0000,    0.3438},
      {0.6719,    1.0000,    0.3281},
      {0.6875,    1.0000,    0.3125},
      {0.7031,    1.0000,    0.2969},
      {0.7188,    1.0000,    0.2813},
      {0.7344,    1.0000,    0.2656},
      {0.7500,    1.0000,    0.2500},
      {0.7656,    1.0000,    0.2344},
      {0.7813,    1.0000,    0.2188},
      {0.7969,    1.0000,    0.2031},
      {0.8125,    1.0000,    0.1875},
      {0.8281,    1.0000,    0.1719},
      {0.8438,    1.0000,    0.1563},
      {0.8594,    1.0000,    0.1406},
      {0.8750,    1.0000,    0.1250},
      {0.8906,    1.0000,    0.1094},
      {0.9063,    1.0000,    0.0938},
      {0.9219,    1.0000,    0.0781},
      {0.9375,    1.0000,    0.0625},
      {0.9531,    1.0000,    0.0469},
      {0.9688,    1.0000,    0.0313},
      {0.9844,    1.0000,    0.0156},
      {1.0000,    1.0000,         0},
      {1.0000,    0.9844,         0},
      {1.0000,    0.9688,         0},
      {1.0000,    0.9531,         0},
      {1.0000,    0.9375,         0},
      {1.0000,    0.9219,         0},
      {1.0000,    0.9063,         0},
      {1.0000,    0.8906,         0},
      {1.0000,    0.8750,         0},
      {1.0000,    0.8594,         0},
      {1.0000,    0.8438,         0},
      {1.0000,    0.8281,         0},
      {1.0000,    0.8125,         0},
      {1.0000,    0.7969,         0},
      {1.0000,    0.7813,         0},
      {1.0000,    0.7656,         0},
      {1.0000,    0.7500,         0},
      {1.0000,    0.7344,         0},
      {1.0000,    0.7188,         0},
      {1.0000,    0.7031,         0},
      {1.0000,    0.6875,         0},
      {1.0000,    0.6719,         0},
      {1.0000,    0.6563,         0},
      {1.0000,    0.6406,         0},
      {1.0000,    0.6250,         0},
      {1.0000,    0.6094,         0},
      {1.0000,    0.5938,         0},
      {1.0000,    0.5781,         0},
      {1.0000,    0.5625,         0},
      {1.0000,    0.5469,         0},
      {1.0000,    0.5313,         0},
      {1.0000,    0.5156,         0},
      {1.0000,    0.5000,         0},
      {1.0000,    0.4844,         0},
      {1.0000,    0.4688,         0},
      {1.0000,    0.4531,         0},
      {1.0000,    0.4375,         0},
      {1.0000,    0.4219,         0},
      {1.0000,    0.4063,         0},
      {1.0000,    0.3906,         0},
      {1.0000,    0.3750,         0},
      {1.0000,    0.3594,         0},
      {1.0000,    0.3438,         0},
      {1.0000,    0.3281,         0},
      {1.0000,    0.3125,         0},
      {1.0000,    0.2969,         0},
      {1.0000,    0.2813,         0},
      {1.0000,    0.2656,         0},
      {1.0000,    0.2500,         0},
      {1.0000,    0.2344,         0},
      {1.0000,    0.2188,         0},
      {1.0000,    0.2031,         0},
      {1.0000,    0.1875,         0},
      {1.0000,    0.1719,         0},
      {1.0000,    0.1563,         0},
      {1.0000,    0.1406,         0},
      {1.0000,    0.1250,         0},
      {1.0000,    0.1094,         0},
      {1.0000,    0.0938,         0},
      {1.0000,    0.0781,         0},
      {1.0000,    0.0625,         0},
      {1.0000,    0.0469,         0},
      {1.0000,    0.0313,         0},
      {1.0000,    0.0156,         0},
      {1.0000,         0,         0},
      {0.9844,         0,         0},
      {0.9688,         0,         0},
      {0.9531,         0,         0},
      {0.9375,         0,         0},
      {0.9219,         0,         0},
      {0.9063,         0,         0},
      {0.8906,         0,         0},
      {0.8750,         0,         0},
      {0.8594,         0,         0},
      {0.8438,         0,         0},
      {0.8281,         0,         0},
      {0.8125,         0,         0},
      {0.7969,         0,         0},
      {0.7813,         0,         0},
      {0.7656,         0,         0},
      {0.7500,         0,         0},
      {0.7344,         0,         0},
      {0.7188,         0,         0},
      {0.7031,         0,         0},
      {0.6875,         0,         0},
      {0.6719,         0,         0},
      {0.6563,         0,         0},
      {0.6406,         0,         0},
      {0.6250,         0,         0},
      {0.6094,         0,         0},
      {0.5938,         0,         0},
      {0.5781,         0,         0},
      {0.5625,         0,         0},
      {0.5469,         0,         0},
      {0.5313,         0,         0},
      {0.5156,         0,         0},
      {0.5000,         0,         0}
    };

    r = jet[index][0];
    g = jet[index][1];
    b = jet[index][2];
    return;
  }

}